/*
 * Copyright 1993, 1996, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */

/* tex_dlists.c
 * Demonstrates how to put multiple textures in display lists.
 *
 * For situations where a texture will not fit entirely into memory,
 * it can be split into two.  The polygon onto which the texture would
 * be applied is also split into two, with the first polygon textured
 * with the first "half" texture and the second polygon getting the 
 * second.
 * 
 * If the minification and magnification filters are set to blend 
 * texels, a seam will show up at the boundary between one texture 
 * and the next unless a border is provided on each texture which 
 * imitates the nearest texels of the adjacent texture map.  
 * 
 * This program draws four contiguous polygons; the first pair have
 * appropriate borders while the second pair have no border and 
 * therefore blend at their edges with the assigned 
 * GL_TEXTURE_BORDER_COLOR. 
 *
 *	<a> key		- toggle animation on/off\n"
 *	<f> key		- toggle linear/nearest filter\n"
 *	Escape key	- exit the program
 */
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

#include <math.h>
#include <stdio.h>

#include "rgbImageFile.h"	/* should be in ../../include */

/*  Function Prototypes  */

GLvoid  initgfx( GLvoid );
GLvoid  animate( GLvoid );
GLvoid  visibility( GLint );
GLvoid  drawScene( GLvoid );
GLvoid  reshape( GLsizei, GLsizei );
GLvoid  keyboard( GLubyte, GLint, GLint );

GLvoid  toggleFilters( GLvoid );

GLvoid  initTexture( GLvoid );

void  printHelp( char *progname );

/* Global Definitions */

#define KEY_ESC	27	/* ascii value for the escape key */

/* checkerboard ImageWidth, ImageHeight, ImageBorder */
#define IMGWIDTH 16
#define IMGHEIGHT 16
#define BORDERSZ 2
/* four GLfloats per RGBA texel */
#define RGBASZ 4

/* Global Variables */

static GLboolean animateFlag = GL_TRUE;

static GLfloat scale_factor = 1.0;

static GLuint texListBase;

void
main( int argc, char *argv[] )
{
	GLsizei		width, height;

	glutInit( &argc, argv );

	/* create a window that is 1/4 the size of the screen */

	width = glutGet( GLUT_SCREEN_WIDTH ); 
	height = glutGet( GLUT_SCREEN_HEIGHT );
	glutInitWindowPosition( width / 4, height / 4 );
	glutInitWindowSize( (width / 2) - 4, height / 2 );
	glutInitDisplayMode( GLUT_RGBA | GLUT_DOUBLE );
	glutCreateWindow( argv[0] );

	initTexture();
	initgfx();

	glutKeyboardFunc( keyboard );
	glutIdleFunc( animate );
	glutVisibilityFunc( visibility );
	glutReshapeFunc( reshape );
	glutDisplayFunc( drawScene ); 

	printHelp( argv[0] );

	glutMainLoop();
}

GLvoid
printHelp( char *progname )
{
	fprintf(stdout, "\n%s - demonstrates textures in display lists\n"
		"<a> key	- toggle animation on/off\n"
		"<f> key	- toggle nearest/linear filter\n"
		"Escape key     - exit the program\n\n",
		progname );
	fprintf(stdout, "Filter is GL_LINEAR\n");
}

GLvoid
initgfx( void )
{
	glClearColor( 0.0, 0.0, 0.0, 1.0 );
}

/* Return a checkerboard texture with width w, height h, border size b, 
 * and c components 
 */
GLfloat *
createTexture( GLsizei w, GLint h, GLint b, GLint c )
{
	GLfloat *image;
	int n, m;

	image = (GLfloat *) malloc((w+b) * (h+b) * c * sizeof(GLfloat));	
	for(n = 0; n < (w+b); n++) {
		for(m = 0; m < (h+b); m++) {
			/* toggle opaque red, black: 1,0,0,1  0,0,0,1 */
			image[c*((n*(h+b))+m)+0] = (n+m)%2?1.0:0.0;
			image[c*((n*(h+b))+m)+1] = 0.0;
			image[c*((n*(h+b))+m)+2] = 0.0;
			image[c*((n*(h+b))+m)+3] = 1.0;
		}
	}
	return image;
}

GLvoid 
initTexture()
{
	GLfloat *image[4];
	GLfloat black[4] = { 0.0, 0.0, 0.0, 1.0 };
	int i;

	glEnable( GL_TEXTURE_2D );
	glEnable( GL_BLEND );
	glBlendFunc( GL_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glTexEnvf( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	/* borders have no effect unless texture wrapping is GL_CLAMP */
	glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

	/* borders have no effect unless texels are being averaged */
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	/* load textures into display lists... */
	texListBase = glGenLists(4);

	/* Create two checkerboard textures IMGWIDTHxIMGHEIGHT w/ borders */
	for(i=0;i<2;i++){
		image[i] = createTexture(IMGWIDTH, IMGHEIGHT, BORDERSZ, RGBASZ);
		glNewList(texListBase+i,GL_COMPILE);
			glTexImage2D(GL_TEXTURE_2D, 0, 4, 
				(IMGWIDTH+BORDERSZ), (IMGWIDTH+BORDERSZ), 1, 
				GL_RGBA, GL_FLOAT, image[i]);
		glEndList();
	}

	/* And two checkerboard textures IMGWIDTHxIMGHEIGHT w/out borders */
	for(i=2;i<4;i++){
		image[i] = createTexture(IMGWIDTH, IMGHEIGHT, 0, RGBASZ);
		glNewList(texListBase+i,GL_COMPILE);
			/* border color is used for averaging edges of 
			 * borderless textures */
			glTexParameterfv(GL_TEXTURE_2D, GL_TEXTURE_BORDER_COLOR, black);
			glTexImage2D(GL_TEXTURE_2D, 0, 4, (IMGWIDTH), (IMGWIDTH), 0, 
					GL_RGBA, GL_FLOAT, image[i]);
		glEndList();
	}
}

GLvoid toggleFilters(GLvoid)
{
	static GLint filter = GL_LINEAR;

	switch (filter) {
	case GL_LINEAR:
		filter = GL_NEAREST;
		fprintf(stdout, "Filter is GL_NEAREST\n");
		break;
	case GL_NEAREST:
		filter = GL_LINEAR;
		fprintf(stdout, "Filter is GL_LINEAR\n");
		break;
	}

	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, filter);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, filter);
}

GLvoid 
keyboard( GLubyte key, GLint x, GLint y )
{
	switch (key) {
	case 'a':	/* toggle animation */
		animateFlag = !animateFlag;
		if (animateFlag)
			glutIdleFunc( animate );
		else
			glutIdleFunc( NULL );
		break;
	case 'f':
		toggleFilters();
		glutPostRedisplay();
		break;
	case KEY_ESC:	/* Exit whenever the Escape key is pressed */
		exit(0);
	}
}

GLvoid 
animate( GLvoid )
{
	scale_factor *= 0.99;
	if (scale_factor < 0.05) scale_factor = 1.0;

	/* Tell GLUT to redraw the scene */
	glutPostRedisplay();
}

GLvoid
visibility( int state ) 
{
	if (state == GLUT_VISIBLE && animateFlag) {
		glutIdleFunc( animate );
	} else {
		glutIdleFunc( NULL );
	}
}

GLvoid
reshape( GLsizei width, GLsizei height )
{
	glViewport( 0, 0, width, height );

	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	glOrtho( 0.0, 6.0, -2.0, 2.0, -2.0, 2.0 );
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
}

GLvoid
drawScene(void)
{
	int n;

	/* texture coordinates are equal to vertex coordinates */
	static float t0[2] = { 0.0, 0.0 };
	static float t1[2] = { 1.0, 0.0 };
	static float t2[2] = { 1.0, 1.0 };
	static float t3[2] = { 0.0, 1.0 };

	glClear( GL_COLOR_BUFFER_BIT );

	glColor4f( 1.0, 1.0, 1.0, 1.0 );
	glPushMatrix ();
		glScalef(scale_factor, scale_factor, 1.0);
		for( n = 0; n < 4; n++ ) {
			glTranslatef(1.0,0.0,0.0);
			glCallList(texListBase+n);
			glBegin( GL_QUADS );
				glTexCoord2fv(t0); glVertex2fv(t0); 
				glTexCoord2fv(t1); glVertex2fv(t1);
				glTexCoord2fv(t2); glVertex2fv(t2);
				glTexCoord2fv(t3); glVertex2fv(t3);
			glEnd();
		}
	glPopMatrix ();
	glutSwapBuffers();
}
