/*
 * Copyright 1996, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */

/* texenv.c 
 * This program reads in a texture from a .rgb file, 
 * and maps it onto a polygon using explicit coordinates.
 *
 * Decal modes uses only the texture color, not the original polygon 
 * color, unless there are alpha values in the texture, in 
 * which case alpha determines the percentage of texture 
 * blended with the polygon color.  Decal may be faster than the 
 * default modulate mode.
 *
 * Modulate mode modulates the color of the polygon by the color
 * of the texture, so a white polygon should be used to get the
 * full texture color.  If alpha is present, it is modulated by
 * the texture alpha.  
 *
 * Blend mode uses the texture color as though it were an alpha
 * value.  It uses the texture color to blend the fragment color
 * with a constant blending color.  If alpha is present, it is 
 * modulated by the texture alpha.
 *
 * Replace mode replaces both the fragments color and alpha values
 * with the corresponding texture values.
 *
 * The texture environment mode and the polygon color can be 
 * changed interactively.
 *
 *	<c> Key		- toggle between white and red polygon
 *	<e> Key		- cycle through environment modes
 *	Escape Key	- exit program
 */
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

#include <math.h>
#include <stdio.h>

#include "rgbImageFile.h"	/* should be in ../../include */

/*  Function Prototypes  */

GLvoid  initgfx( GLvoid );
GLvoid  drawScene( GLvoid );
GLvoid  reshape( GLsizei, GLsizei );
GLvoid  keyboard( GLubyte, GLint, GLint );

GLvoid  initTexture( unsigned int *, GLsizei, GLsizei );

void resetView( GLvoid );
void printHelp( char * );

/* Global Definitions */

#define KEY_ESC	27	/* ascii value for the escape key */

/* Global Variables */

static GLuint 		envmode = 0;

static GLfloat		white[] = { 1.0, 1.0, 1.0, 1.0 };
static GLfloat		red[] = { 1.0, 0.0, 0.0, 1.0 };
static GLfloat		water[] = { 0.0, 0.3, 0.5, 1.0 };

static GLboolean 	useWhite = GL_FALSE;

typedef struct {
	GLint	type;
	char	*name;
} EnvModeInfo;

static EnvModeInfo modes[4] = {
	{ GL_DECAL, "GL_DECAL" },
	{ GL_MODULATE, "GL_MODULATE" },
	{ GL_BLEND, "GL_BLEND" },
	{ GL_REPLACE_EXT, "GL_REPLACE_EXT" },
};

void
main ( int argc, char *argv[])
{
	GLsizei 	width, height;
	char		*imageFileName = "fish.rgba";
	unsigned int 	*image;
	GLsizei		imageWidth, imageHeight;
	
	glutInit( &argc, argv );

	if (argc < 2) {
		fprintf (stderr, "usage: %s <imageFileName>\n", argv[0] );
	} else
		imageFileName = argv[1];

	fprintf(stdout, "using image %s\n\n", imageFileName );

	image = rgbReadImageFile(imageFileName, &imageWidth,
			 &imageHeight);

	/* create a window that is 1/4 the size of the screen */

	width = glutGet( GLUT_SCREEN_WIDTH ); 
	height = glutGet( GLUT_SCREEN_HEIGHT );
	glutInitWindowPosition( (width / 2) + 4, height / 4 );
	glutInitWindowSize( (width / 2) - 4, height / 2 );
	glutInitDisplayMode( GLUT_RGBA | GLUT_DOUBLE );
	glutCreateWindow( argv[0] );

	initTexture( image, imageWidth, imageHeight );
	initgfx();

	glutKeyboardFunc( keyboard );
	glutReshapeFunc( reshape );
	glutDisplayFunc( drawScene ); 

	printHelp( argv[0] );

	glutMainLoop();
}

void
printHelp( char *progname )
{
	fprintf(stdout, "\n%s - demonstrates texture environment modes\n\n" 
		"<c> Key	- toggle polygon color between white/red \n"
		"<e> Key	- cycle through texture environment modes\n"
		"Escape Key	- exit the program\n\n",
		progname);
	printf("\nTexture Environment Mode is %s\n", modes[envmode].name);
}

GLvoid 
initgfx()
{
	glClearColor( water[0], water[1], water[2], water[3] );
}

GLvoid 
initTexture( unsigned int *image, 
	GLsizei imageWidth, GLsizei imageHeight )
{
	/* use gluBuild2DMipmaps to create and load mipmaps (it will 
	 * also scale he original image to be a power of two, if 
	 * necessary)
	 *
	 *      gluBuild2DMipmaps( target, components, width, height,
	 *              format,  type, imageArray ) 
	 */

	gluBuild2DMipmaps( GL_TEXTURE_2D, 4, imageWidth, imageHeight,
			 GL_RGBA, GL_UNSIGNED_BYTE, image );

	/* Setting the magnification filter to nearest instead of linear 
	 * may run faster on some platforms, with possibly lower quality 
	 */
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, modes[envmode].type);

	/* Used when tex env mode is GL_BLEND */
	glTexEnvfv( GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, water );

	/* enable 2D texture mapping */
	glEnable( GL_TEXTURE_2D );
}

GLvoid
reshape( GLsizei width, GLsizei height )
{
	GLdouble      aspect;

	glViewport( 0, 0, width, height );

	aspect = (GLdouble) width / (GLdouble) height;

	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluPerspective( 45.0, aspect, 1.0, 50.0 );
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	glTranslatef( 0.0, 0.0, -12.0 ); 
}

GLvoid
cycleTexEnvMode( GLvoid )
{
	envmode = (envmode + 1) % 4;
	
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, modes[envmode].type);
	printf("Texture Environment Mode is %s\n", modes[envmode].name );
}

GLvoid 
keyboard( GLubyte key, GLint x, GLint y )
{
	switch (key) {
	case 'c':	/* toggle polygon color */
		useWhite = !useWhite;
		glutPostRedisplay();
		break;
	case 'e':	/* cycle through texture environment modes */
		cycleTexEnvMode();
		glutPostRedisplay();
		break;
	case KEY_ESC:	/* Exit whenever the Escape key is pressed */
		exit(0);
	}
}

GLvoid
drawScene(void)
{
	static float v0[3] = { -1.5, -1.0, 0.0 };
	static float v1[3] = {  1.5, -1.0, 0.0 };
	static float v2[3] = {  1.5,  1.0, 0.0 };
	static float v3[3] = { -1.5,  1.0, 0.0 };

	static float t0[2] = { 0.0, 0.0 };
	static float t1[2] = { 1.0, 0.0 };
	static float t2[2] = { 1.0, 1.0 };
	static float t3[2] = { 0.0, 1.0 };

	glClear( GL_COLOR_BUFFER_BIT );

	if (useWhite) {
		glColor4fv( white );
	} else {
		/* Make the polygon RED so that we can see what happens
		 * in modulate and blend mode; usually, you would make
		 * the polygon white.
		 */ 
		glColor4fv( red );
	} 
	glPushMatrix ();

		glBegin( GL_QUADS );
			glTexCoord2fv( t0 ); glVertex3fv( v0 );
			glTexCoord2fv( t1 ); glVertex3fv( v1 );
			glTexCoord2fv( t2 ); glVertex3fv( v2 );
			glTexCoord2fv( t3 ); glVertex3fv( v3 );
		glEnd();

	glPopMatrix ();

	glutSwapBuffers();
}
